from email.message import EmailMessage
import os
import smtplib

from utilspkg import utils_init

if __name__ == '__main__':
    utils_init.load_env_variables_from_yaml('/Users/croft/VScode/ptagit/env_vars.yaml')

logger = utils_init.setup_logger(__name__)

class EmailConnect:
    """
    A utility class to send emails using Gmail's SMTP service.
    """

    def __init__(self, user_name=None, pw=None, from_name=None, testing_flag=False, testing_email=None):
        """
        Constructs an instance of the EmailSender class.
        
        Args:
            user_name (str, optional): Gmail username to send emails. Defaults to GMAIL_USER environment variable.
            pw (str, optional): Gmail password. Defaults to GMAIL_PASSWORD environment variable.
            from_name (str, optional): Sender's display name in email. Defaults to GMAIL_FROM environment variable.
            testing_flag (bool, optional): Indicates whether to use a testing email address. Defaults to False.
            testing_email (str, optional): Testing email address. Defaults to TESTING_EMAIL environment variable.
        """
        self.user_name = user_name if user_name else os.environ['GMAIL_USER']
        self.pw = pw if pw else os.environ['GMAIL_PASSWORD']
        self.from_name = from_name if from_name else os.environ['GMAIL_FROM']
        self.testing_flag = testing_flag
        self.testing_email = testing_email if testing_email else os.environ['TESTING_EMAIL']


    def send_email(self, to, subject, body, testing_flag=None):
        """
        Sends an email with the specified details using the Gmail SMTP service.

        Args:
            to (str): The recipient's email address.
            subject (str): The email subject.
            body (str): The email body.
            testing_flag (bool, optional): If set, overrides the `testing_flag` attribute of the class. Defaults to None.
        """
        if testing_flag is None:
            testing_flag = self.testing_flag
        to = to if not testing_flag else self.testing_email
        message = EmailMessage()
        message.set_content(body)
        message["From"] = self.from_name
        message["To"] = to
        message["Subject"] = subject

        with smtplib.SMTP_SSL("smtp.gmail.com", 465) as server:
            server.login(self.user_name, self.pw)
            server.send_message(message)



import os
import time
import logging
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

from utilspkg import utils_init

if __name__ == '__main__':
    utils_init.load_env_variables_from_yaml('/Users/croft/VScode/ptagit/env_vars.yaml')

# logger = utils_init.setup_logger(__name__)

class SlackConnect:
    def __init__(self, api_key=None, testing_flag=False, logger=None):
        """
        Initialize the SlackSender with an API key, testing flag, and optional logger.
        """
        # get the value of 'SLACK_API_KEY' from os if it exists
        self.SLACK_API_KEY = os.getenv("SLACK_API_KEY")

        # handle the old variable name for a while
        if not self.SLACK_API_KEY:
            self.SLACK_API_KEY = os.environ["SLACK_ACCESS_TOKEN_TEAM"]

        # I'm retiring TESTING_DM in favor of TESTING_CHANNEL, but handling it in the interim
        self.TESTING_CHANNEL = os.getenv ("TESTING_DM")

        # handle the old variable name for a while
        if not self.TESTING_CHANNEL:
            self.TESTING_CHANNEL = os.environ["TESTING_CHANNEL"] # not currently used since all goes to the DM's "channel"

        self.api_key = api_key if api_key else self.SLACK_API_KEY
        self.slack_client = WebClient(token=self.api_key)
        self.logger = logger if logger else utils_init.setup_logger(__name__)  #logging.getLogger(__name__)
        self.testing_flag = testing_flag
        self.testing_dm_or_channel = self.TESTING_CHANNEL


    def make_slack_api_call(self, api_method, **kwargs):
        while True:
            try:
                response = api_method(**kwargs)
                return response
            except SlackApiError as e:
                if e.response["error"] == "ratelimited":
                    delay = int(e.response.headers.get('Retry-After'))
                    time.sleep(delay)
                else:
                    self.print(f"Error making API call: {e}")
                    raise e
    
    def send_dm_or_channel_message(self, channel_or_slack_id, message, thread_ts=None, testing_flag=None):
        """
        Send a message to the given channel or user (by Slack ID) through Direct Message.
        Optionally, reply to a threaded message by providing a thread timestamp.
        """
        channel_or_slack_id = channel_or_slack_id if not testing_flag else self.testing_dm_or_channel

        if channel_or_slack_id.startswith("U"):
            dm = self.make_slack_api_call(self.slack_client.conversations_open, users=channel_or_slack_id)
            channel_or_slack_id = dm['channel']['id']

        response = {}

        if not message:
            self.print(f"Message is empty")
            response['ok'] = False
            response['error'] = "ERROR: Tried to send empty message on Slack via slackutils"
        else:
            response = self.make_slack_api_call(
                self.slack_client.chat_postMessage,
                channel=channel_or_slack_id,
                text=message,
                thread_ts=thread_ts,
                unfurl_links=False,
                unfurl_media=False
            )

        return response
    
    def get_list_of_channels(self, bool_public_channels=True, bool_private_channels=True, exclude_archived=True):
        '''https://api.slack.com/methods/conversations.list
            Returns the conversations_list() result of 'all channels
            The object can then be iterated over (for channel in all_channels). Some properties:
            channel_id = channel["id"]
            channel_name = channel["name"]
            status = "private" if channel["is_private"] else "public"
            description = channel["purpose"]["value"]
            topic = channel["topic"]["value"]
            is_archived = channel["is_archived"]
        '''
        public_channels = None
        private_channels = None
        # Get the public channels
        if bool_public_channels:
            response = self.make_slack_api_call(self.slack_client.conversations_list, types="public_channel", exclude_archived=exclude_archived)
            public_channels = response["channels"]

        # Get the private channels
        if bool_private_channels:
            response = self.make_slack_api_call(self.slack_client.conversations_list, types="private_channel", exclude_archived=exclude_archived)
            private_channels = response["channels"]

        all_channels = public_channels + private_channels if public_channels and private_channels else public_channels or private_channels
        
        return all_channels

    def get_channel_messages(self, channel_id, oldest_timestamp=None, newest_timestamp=None):
        """
        Fetch messages from the specified Slack channel.
        Optionally, filter the messages by providing the oldest and/or newest timestamp.
        Handles pagination and rate limits.
        """
        messages = []
        next_cursor = None

        while True:
            # Build kwargs dict based on parameters
            kwargs = {
                "channel": channel_id,
                "inclusive": False,
                "limit": 100,  # Get maximum number of messages per API call
                "cursor": next_cursor,
            }

            # Only add timestamp parameters if they are not None
            if oldest_timestamp:
                kwargs["oldest"] = oldest_timestamp
            if newest_timestamp:
                kwargs["latest"] = newest_timestamp

            # Request the conversation history using make_slack_api_call
            response = self.make_slack_api_call(self.slack_client.conversations_history, **kwargs)
            messages += response.data.get('messages')

            # Check if more messages are available
            next_cursor = response.data.get('response_metadata', {}).get('next_cursor')
            if not next_cursor:
                break

            # Pause before next API call to avoid hitting rate limits (if needed)
            time.sleep(1)

        return messages

  
    def get_channel_members(self, channel_id):
        '''Takes a channel_id string (e.g. "C0921V92") and returns the results of slack_client.conversation_members()'''
        return self.slack_client.conversations_members(channel=channel_id)


    def get_users (self):
        '''return the result of slack_client.users_list()'''
        return self.make_slack_api_call(self.slack_client.users_list)
        


if __name__ == '__main__':
    slack = SlackConnect()
    channels = slack.get_list_of_channels()
    print (len(channels))


import os
from datetime import datetime, timedelta
import pytz

import json
from slack_sdk.errors import SlackApiError

# do all of my standard imports to start a script
from utilspkg import utils_init

# creates a logger
logger = utils_init.setup_logger(__name__)

# loads ENV file when running locally
if __name__ == '__main__':
    utils_init.load_env_variables_from_yaml('/Users/croft/VScode/ptagit/env_vars.yaml')

# my custom Slack, DB, GPT, etc connections file
from utilspkg import utils_connections #, utils_times

# creates instances of my Slack, DB, GPT, etc classes
db, slack, gpt = utils_connections.get_connections()

EXCLUDED_SLACK_IDS = ["U059UA4RPHB", #inna / cam 
                      "USLACKBOT", #slackbot
                      "U059UA4K953", #aron
                      "U05ALUWSL00", #kat
                      "U05A8HMUGKF", #welcome bot
                      "U059YQXHBCN" # troy 'bot' probably delete
                      , "U059X0WFVST" #troy 
                      ]

# DB Tables
tasks_table = 'Tasks'
messages_table = 'Messages' 
channels_table = 'Slack Channels'
students_table = 'Students'
slack_users_table = 'Slack Users'
all_messages_table = 'All Messages' 
profile_pic_changed_table = 'OB-ProfilePicChangedUsers'
introduction_posted_table = 'OB-IntroductionPostedUsers'
onboarding_done_table = 'OB-OnboardingDoneUsers'


def add_slack_id_to_db(user_id):
    #db, slack, gpt = utils_connections.get_connections()

    # Query Slack API to fetch user's profile data
    result = slack.slack_client.users_info(user=user_id)
    new_user = result['user']
    return add_slack_user_to_db(new_user)


def add_slack_user_to_db (user):
    #db, slack, gpt = utils_connections.get_connections()


    profile = user.get('profile', {})
    email = profile.get('email', '')
    is_email_confirmed = user.get('is_email_confirmed', False)
    slack_id = user.get('id', '')
    real_name = user.get('real_name', '')
    tz_offset = user.get('tz_offset')
    tz = user.get('tz')
    raw_data = json.dumps(user)

    # Updated fields
    is_bot = user.get('is_bot', False)
    is_bot = True if slack_id in EXCLUDED_SLACK_IDS else is_bot
    custom_image = profile.get('is_custom_image', False)
    deactivated = user.get('deleted', False)

    link_to_student_table = get_student_link(email=email) if email else [] 

    fields = {
        'Email': email,
        'Student': link_to_student_table,
        'Confirmed Email?': is_email_confirmed,
        'Slack ID': slack_id,
        'Real Name': real_name,
        'Timezone': tz,
        'TZ Offset': tz_offset,
        'Raw Data': raw_data,
        'Ignore': is_bot,  # New field: checkbox true if user is a bot
        'Custom Profile Pic': custom_image,  # New field: checkbox true if custom image is true
        'Deactivated': deactivated,  # New field: checkbox true if "deleted" is true
    }

    # Check if record already exists
    records = db.search_table(slack_users_table, 'Slack ID', slack_id)

    if records:
        # Record exists, update the existing record
        record_id = records[0]['id']
        try:
            record = db.update_record(slack_users_table, record_id, fields)
        except Exception as e:
            print(f"Error updating record: {e}")
    else:
        # Record not found, insert new record
        try:
            record = db.insert_record(slack_users_table, fields)
        except Exception as e:
            print(f"Error inserting record: {e}")
    
    return record
    

def get_challenge_start_date(optional_date=None, optional_timezone='US/Eastern'):
    """
    Unless specified, uses NOW() and US/Eastern. Returns the MONDAY start date of the current challenge as a date-only datetime object. 
    
    Optional args: 
    - date (defaults to today)
    - timezone (defaults to US/Eastern)
    """
    # in case i accidentally override the default but send nothing
    if optional_timezone is None:
        optional_timezone='US/Eastern'

    timezone = pytz.timezone(optional_timezone)

    if optional_date is None:
        optional_date = datetime.now(timezone)

    # if its Sunday, return today's date. Otherwise, return the previous Sunday's date
    if optional_date.weekday() == 6:
        challenge_start_date = optional_date
    else:
        challenge_start_date = optional_date - timedelta(days=optional_date.weekday()+1)
    
    # adjust to the Monday (1 day later)
    challenge_start_date = challenge_start_date + timedelta(days=1)

    # return the date-only datetime object
    challenge_start_date = datetime(challenge_start_date.year, challenge_start_date.month, challenge_start_date.day)

    return challenge_start_date


def get_student_link(email='', slack_id=None):
    #db, slack, gpt = utils_connections.get_connections()
    filter_formula = ''
    student_records = []
    # get BOT_ALERTS_CHANNEL from env. if not available, get TESTING_CHANNEL
    bot_alerts_channel = os.getenv("BOT_ALERTS_CHANNEL")
    if not bot_alerts_channel:
        bot_alerts_channel = os.environ["TESTING_CHANNEL"]
        if not bot_alerts_channel:
            print("BOT_ALERTS_CHANNEL not found in env variables. Please add it to your env variables or to the env_vars.yaml file.")
            # exit(1)

    # if we have email and it's a valid email
    if email and email.find("@") > 0 :
        filter_formula = f'FIND("{email}", {{Email - Search}})'
        # print(f"Formula in get_student_link(): {filter_formula}")
    elif slack_id:
        filter_formula = f'FIND("{slack_id}", {{Slack ID}})'

    if filter_formula: #don't search if we don't have a valid filter formula
        student_records = db.get_records(students_table, formula=filter_formula)

    if len(student_records) == 0:
        # print(f"No student found! {email}")
        link_to_student = []
    else:
        link_to_student = [student_records[0]['id']]
        # print(f"CHALLENGER FOUND: {link_to_student}. {email}")

    # SEND MYSELF SOME WARNING MESSAGES 
    if not filter_formula:
        # that's weird. alert me
        message = f"WEIRD: No email and no slack_id sent to get_student_link just now"
        slack.send_dm_or_channel_message (bot_alerts_channel, message)

    if len(student_records) > 1:
        print(f"WARNING: Found more than one challenger with email {email}")
        message = f"WARNING: Found more than one challenger with email {email}"
        slack.send_dm_or_channel_message (bot_alerts_channel, message)

    return link_to_student




def add_message_to_db (data, table=messages_table):
    #db, slack, gpt = utils_connections.get_connections()

    
    try:
        TESTING_FLAG = os.getenv("TESTING_FLAG")
        logger.debug(f"IN 'add message to db")
        
        event = data #.get('event', {})
        
        # Check if the user is in the exclusion list or if no user exists!
        if not event.get('user'):
            print(f"SKIPPING: NO USER FOUND.")
            return None
        
        if not TESTING_FLAG and (event.get('user') in EXCLUDED_SLACK_IDS or event.get('user') == 'USLACKBOT' or event.get('bot_id')):
            print(f"SKIPPING: MESSAGE FROM EXCLUDED BOT/USER.")
            return None
        
        print(f"Processing message from {event.get('user')}...")
        
        ts = event.get('ts')
        channel = event.get('channel')
        slack_id = event.get('user')
        print(f"checking for existing message with ts: {ts}")
        # Check if the message with the same ts already exists in the Messages table
        existing_message = db.search_table(table, 'Timestamp', ts)

        if not existing_message:
            # Find the corresponding channel and student records
            channel_records = db.search_table(channels_table, 'Channel ID', channel)
            channel_record = channel_records[0] if channel_records else None

            # student_records = db.search_table(students_table, 'Slack ID', event.get('user'))
            # student_record = student_records[0] if student_records else None
            slack_records = db.search_table(slack_users_table, 'Slack ID', slack_id)
            slack_record = slack_records[0] if slack_records else None
            slack_timezone = slack_record['fields'].get('Timezone','') if slack_record else None

            student_record = get_student_link (slack_id=slack_id)
            # Call the chat.getPermalink method using the slackclient (part of slack object)
            try:
                result = slack.slack_client.chat_getPermalink(channel=channel, message_ts=ts)
                permalink = result.get('permalink')
            except SlackApiError as e:
                permalink = ''
                # You will get a SlackApiError if "ok" is False
                assert e.response["ok"] is False
                assert e.response["error"]  # str like 'invalid_auth', 'channel_not_found'
                print(f"Got an error: {e.response['error']}")
            
            # set the has files veraible if files are attached
            has_files = True if event.get('files') else False

            # need to add the challenge date to the message record
            challenge_start_date = get_challenge_start_date(optional_timezone=slack_timezone)
            print(f"Calculated challenge start date: {challenge_start_date}")

            # but need to take into account that messages posted in goal for the week and introduce yourself should be the following week if they are posted on Thursday or later
            if channel_record and channel_record['fields'].get('Channel Name') in ['-goal-for-the-week', '--introduce-yourself']:
                if datetime.now().weekday() >= 3:
                    challenge_start_date = challenge_start_date + timedelta(days=7)
                    print(f"Adjusted challenge start date: {challenge_start_date}")

            # formatting for airtable iso format, date only
            challenge_start_date = challenge_start_date.date().isoformat()
            # Add a new row to the Messages table
            new_message = {
                'Channel ID': channel,
                'Channel': [channel_record['id']] if channel_record else [],
                'Timestamp': ts,
                'Challenge Date': challenge_start_date,
                'Raw Data': json.dumps(event),
                'Message Link': permalink,
                'Message Text': event.get('text'),
                'Slack User': [slack_record['id']] if slack_record else [],
                'Has Files': has_files,
                'Student': student_record,
            }
            #print(f"Airtable array to insert: {new_message}")
            new_record = db.insert_record(table, new_message)
            # new_record_id = new_record['id']
            print("SUCCESS! Written to airtable")
            return new_record
        # if the message is already in our database, return 'true' so pub/sub acknowledges (dismisses) it
        else:
            print("SKIPPING: MESSAGE ALREADY IN OUR DATABASE!")
            return None
        
    except Exception as e:
        print(f"**utils_pta.py - write_to_airtable fx error writing to Airtable: {e}")
        raise




import os
# from setuptools import setup
from setuptools import setup, find_packages

os.chdir('/Users/croft/VScode/ptacorechallenge/utilsfolder')

setup(
    name='utilspkg',
    version='0.4.8',
    packages=find_packages(include=['utilspkg']),
    install_requires=[
        'airtable-python-wrapper>=0.15.3',
        'openai>=0.27.7',
        'protobuf>=4.24.1',
        'python-dotenv>=1.0.0',
        'pytz>=2023.3',
        'slack-sdk>=3.21.3',
        'tenacity>=8.2.2',
        'PyYAML',
        'google-cloud-logging'
    ]
)
